@@ -2,7 +2,7 @@
  * Basic partition manager for the installer.
  ***************************************************************************
  *
- *   Copyright (C) 2019, 2020-2021 by AK-47
+ *   Copyright (C) 2019, 2020-2022 by AK-47
  *   Transplanted code, marked with comments further down this file:
  *    - Copyright (C) 2003-2010 by Warren Woodford
  *    - Heavily edited, with permision, by anticapitalista for antiX 2011-2014.
@@ -22,11 +22,11 @@
  * This file is part of the gazelle-installer.
  ***************************************************************************/
 
-#include <sys/sysinfo.h>
 #include <sys/stat.h>
 #include <QDebug>
 #include <QLocale>
 #include <QFileInfo>
+#include <QDir>
 #include <QJsonDocument>
 #include <QJsonObject>
 #include <QJsonArray>
@@ -134,8 +134,8 @@ void PartMan::scan(DeviceItem *drvstart)
         drvit->device = jdName;
         drvit->path = jdPath;
         const QString &ptType = jsonDrive["pttype"].toString();
-        drvit->flags.curGPT = (ptType=="gpt");
-        drvit->flags.curMBR = (ptType=="dos");
+        drvit->flags.curGPT = (ptType == "gpt");
+        drvit->flags.curMBR = (ptType == "dos");
         drvit->flags.rotational = jsonDrive["rota"].toBool();
         drvit->discgran = jsonDrive["disc-gran"].toInt();
         drvit->size = jsonDrive["size"].toVariant().toLongLong();
@@ -152,6 +152,7 @@ void PartMan::scan(DeviceItem *drvstart)
             partit->flags.oldLayout = true;
             partit->device = jsonPart["name"].toString();
             partit->path = jsonPart["path"].toString();
+            partit->uuid = jsonPart["uuid"].toString();
             partit->order = order.indexOf(partit->device);
             partit->size = jsonPart["size"].toVariant().toLongLong();
             partit->physec = jsonPart["phy-sec"].toInt();
@@ -163,8 +164,9 @@ void PartMan::scan(DeviceItem *drvstart)
             if ((partflags & 0x80) || (partflags & 0x04)) partit->setActive(true);
             partit->mapCount = jsonPart["children"].toArray().count();
             partit->flags.curESP = partTypeName.startsWith("EFI "); // "System"/"(FAT-12/16/32)"
-            partit->flags.bootRoot = (!bootUUID.isEmpty() && jsonPart["uuid"]==bootUUID);
+            partit->flags.bootRoot = (!bootUUID.isEmpty() && partit->uuid == bootUUID);
             partit->curFormat = jsonPart["fstype"].toString();
+            if (partTypeName == "BIOS boot") partit->curFormat = "BIOS-GRUB";
             // Touching MacOS first drive, or MS LDM may brick the system.
             if ((proc.detectMac() && partit->device.startsWith("sda"))
                 || partTypeName.startsWith("Microsoft LDM")) {
@@ -253,11 +255,12 @@ void PartMan::scanVirtualDevices(bool rescan)
             devit = new DeviceItem(DeviceItem::VirtualBD, vdlit);
             devit->device = jsonDev["name"].toString();
             devit->path = path;
+            devit->uuid = jsonDev["uuid"].toString();
             devit->flags.rotational = rota;
             devit->discgran = discgran;
             devit->size = size;
             devit->physec = physec;
-            devit->flags.bootRoot = (!bootUUID.isEmpty() && jsonDev["uuid"]==bootUUID);
+            devit->flags.bootRoot = (!bootUUID.isEmpty() && devit->uuid == bootUUID);
             devit->curLabel = label;
             devit->curFormat = jsonDev["fstype"].toString();
             devit->flags.volCrypto = crypto;
@@ -613,9 +616,9 @@ void PartMan::partMenuUnlock(DeviceItem *twit)
     if (dialog.exec() == QDialog::Accepted) {
         qApp->setOverrideCursor(Qt::WaitCursor);
         gui.boxMain->setEnabled(false);
-        const QString &mapdev = editVDev->text();
-        const bool ok = luksOpen(twit, mapdev, editPass->text().toUtf8());
-        if (ok) {
+        try {
+            const QString &mapdev = editVDev->text();
+            luksOpen(twit, mapdev, editPass->text().toUtf8());
             twit->usefor.clear();
             twit->addToCrypttab = checkCrypttab->isChecked();
             twit->mapCount++;
@@ -623,10 +626,13 @@ void PartMan::partMenuUnlock(DeviceItem *twit)
             scanVirtualDevices(true);
             resizeColumnsToFit();
             treeSelChange();
-        }
-        gui.boxMain->setEnabled(true);
-        qApp->restoreOverrideCursor();
-        if (!ok) {
+            gui.boxMain->setEnabled(true);
+            qApp->restoreOverrideCursor();
+        } catch(const char *msg) {
+            // This time, failure is not a dealbreaker.
+            qDebug() << "partMenuUnlock() failed:" << msg;
+            gui.boxMain->setEnabled(true);
+            qApp->restoreOverrideCursor();
             QMessageBox::warning(master, QString(),
                 tr("Could not unlock device. Possible incorrect password."));
         }
@@ -673,7 +679,7 @@ void PartMan::scanSubvolumes(DeviceItem *partit)
     proc.exec("btrfs", {"subvolume", "list", "/mnt/btrfs-scratch"}, nullptr, true);
     lines = proc.readOutLines();
     proc.exec("umount", {"/mnt/btrfs-scratch"});
-    for (const QString &line : lines) {
+    for (const QString &line : qAsConst(lines)) {
         const int start = line.indexOf("path") + 5;
         if (line.length() <= start) goto END;
         DeviceItem *svit = new DeviceItem(DeviceItem::Subvolume, partit);
@@ -735,7 +741,7 @@ bool PartMan::composeValidate(bool automatic, const QString &project)
             QMessageBox::critical(master, QString(), tr("%1 is already"
                 " selected for: %2").arg(twit->shownDevice(), twit->shownUseFor()));
             return false;
-        } else if(!mount.isEmpty() && mount != "FORMAT" && mount != "ESP"){
+        } else if(!mount.isEmpty() && mount != "FORMAT" && mount != "ESP" && mount != "BIOS-GRUB"){
             mounts.insert(mount, *it);
         }
         if (item->type != DeviceItem::VirtualBD) {
@@ -790,8 +796,11 @@ bool PartMan::composeValidate(bool automatic, const QString &project)
                 const QString &use = partit->realUseFor();
                 QString actmsg;
                 if (drvit->flags.oldLayout) {
-                    if (use.isEmpty()) continue;
-                    else if (partit->willFormat()) actmsg = tr("Format %1 to use for %2");
+                    if (use.isEmpty()) {
+                        if (partit->curFormat == "BIOS-GRUB") hasBiosGrub = true;
+                        continue;
+                    }
+                    if (partit->willFormat()) actmsg = tr("Format %1 to use for %2");
                     else if (use != "/") actmsg = tr("Reuse (no reformat) %1 as %2");
                     else actmsg = tr("Delete the data on %1 except for /home, to use for %2");
                 } else {
@@ -886,20 +895,23 @@ bool PartMan::checkTargetDrivesOK()
     return true;
 }
 
-bool PartMan::luksFormat(DeviceItem *partit, const QByteArray &password)
+void PartMan::luksFormat(DeviceItem *partit, const QByteArray &password)
 {
     QStringList cargs({"--batch-mode", "--key-size=512", "--hash=sha512", "--pbkdf=argon2id"});
     if (partit->physec > 0) cargs.append("--sector-size=" + QString::number(partit->physec));
     cargs << "luksFormat" << partit->path;
-    if (!proc.exec("cryptsetup", cargs, &password)) return false;
-    return true;
+    if (!proc.exec("cryptsetup", cargs, &password)) {
+        throw QT_TR_NOOP("Failed to format LUKS container.");
+    }
 }
 
-bool PartMan::luksOpen(DeviceItem *partit, const QString &luksfs, const QByteArray &password)
+void PartMan::luksOpen(DeviceItem *partit, const QString &luksfs, const QByteArray &password)
 {
     QStringList cargs({"open", partit->path, luksfs, "--type", "luks"});
     if (partit->discgran) cargs.prepend("--allow-discards");
-    return proc.exec("cryptsetup", cargs, &password);
+    if (!proc.exec("cryptsetup", cargs, &password)) {
+        throw QT_TR_NOOP("Failed to open LUKS container.");
+    }
 }
 
 DeviceItem *PartMan::selectedDriveAuto()
@@ -944,10 +956,38 @@ int PartMan::countPrepSteps()
     }
     return nstep;
 }
-
-bool PartMan::preparePartitions()
+void PartMan::prepStorage()
 {
-    proc.log(__PRETTY_FUNCTION__);
+    preparePartitions();
+    formatPartitions();
+    mountPartitions();
+    // Run blkid -c /dev/null to freshen UUID cache
+    proc.exec("blkid", {"-c", "/dev/null"});
+    // Refresh the UUIDs
+    proc.exec("lsblk", {"--list", "-bJo", "PATH,UUID"}, nullptr, true);
+    const QJsonObject &jsonObjBD = QJsonDocument::fromJson(proc.readOut(true).toUtf8()).object();
+    const QJsonArray &jsonBD = jsonObjBD["blockdevices"].toArray();
+    QMap<QString, DeviceItem *> alldevs;
+    for (DeviceItemIterator it(*this); DeviceItem *item = *it; it.next()) {
+        alldevs.insert(item->path, item);
+    }
+    for (const QJsonValue &jsonDev : jsonBD) {
+        DeviceItem *devit = alldevs.value(jsonDev["path"].toString());
+        if (devit) devit->uuid = jsonDev["uuid"].toString();
+    }
+}
+void PartMan::installTabs()
+{
+    makeFstab();
+    if (!fixCryptoSetup()) {
+        throw QT_TR_NOOP("Failed to finalize encryption setup.");
+    }
+}
+
+void PartMan::preparePartitions()
+{
+    proc.log(__PRETTY_FUNCTION__, MProcess::Section);
+    static const char *const msgfail = QT_TR_NOOP("Failed to prepare required partitions.");
 
     // Detach all existing partitions.
     QStringList listToUnmount;
@@ -959,7 +999,7 @@ bool PartMan::preparePartitions()
             // Using the existing layout, so only mark used partitions for unmounting.
             for (int ixPart=0; ixPart < partCount; ++ixPart) {
                 DeviceItem *twit = drvit->child(ixPart);
-                if (!twit->usefor.isEmpty()) listToUnmount << twit->device;
+                if (!twit->usefor.isEmpty()) listToUnmount << twit->path;
             }
         } else {
             // Clearing the drive, so mark all partitions on the drive for unmounting.
@@ -967,9 +1007,11 @@ bool PartMan::preparePartitions()
             listToUnmount << proc.readOutLines();
         }
     }
-    for (const QString &devpath : listToUnmount) {
-        proc.exec("swapoff", {devpath});
-        proc.exec("/bin/umount", {devpath});
+    proc.exec("swapon", {"--show=NAME", "--noheadings"}, nullptr, true);
+    const QStringList swaps = proc.readOutLines();
+    for (const QString &devpath : qAsConst(listToUnmount)) {
+        if (swaps.contains(devpath)) proc.shell("grep -q " + devpath + " /proc/swaps && swapoff " + devpath);
+        else proc.exec("/usr/bin/umount", {"-q", devpath});
     }
 
     // Prepare partition tables on devices which will have a new layout.
@@ -988,7 +1030,7 @@ bool PartMan::preparePartitions()
         cargs.append("seek=" + QString::number(offset));
         proc.exec("dd", cargs);
         if (!proc.exec("parted", {"-s", drvit->path, "mklabel",
-            (drvit->willUseGPT() ? "gpt" : "msdos")})) return false;
+            (drvit->willUseGPT() ? "gpt" : "msdos")})) throw msgfail;
     }
 
     // Prepare partition tables, creating tables and partitions when needed.
@@ -1011,7 +1053,7 @@ bool PartMan::preparePartitions()
                 if (useFor.isEmpty()) continue;
                 else if (useFor == "ESP") ptype = useGPT ? "ef00" : "ef";
                 const QStringList &devsplit = DeviceItem::split(twit->device);
-                if (!proc.shell(cmd.arg(devsplit.at(0), devsplit.at(1), ptype))) return false;
+                if (!proc.shell(cmd.arg(devsplit.at(0), devsplit.at(1), ptype))) throw msgfail;
                 proc.status();
             }
         } else {
@@ -1022,7 +1064,7 @@ bool PartMan::preparePartitions()
                 const long long end = start + twit->size / 1048576;
                 const bool rc = proc.exec("parted", {"-s", "--align", "optimal", drvit->path,
                     "mkpart" , "primary", QString::number(start) + "MiB", QString::number(end) + "MiB"});
-                if (!rc) return false;
+                if (!rc) throw msgfail;
                 start = end;
                 proc.status();
             }
@@ -1036,18 +1078,17 @@ bool PartMan::preparePartitions()
                 QStringList cargs({"-s", "/dev/" + devsplit.at(0), "set", devsplit.at(1)});
                 cargs.append(useGPT ? "legacy_boot" : "boot");
                 cargs.append("on");
-                if(!proc.exec("parted", cargs)) return false;
+                if(!proc.exec("parted", cargs)) throw msgfail;
             }
             proc.status();
         }
     }
     proc.exec("partprobe", {"-s"});
-    return true;
 }
 
-bool PartMan::formatPartitions()
+void PartMan::formatPartitions()
 {
-    proc.log(__PRETTY_FUNCTION__);
+    proc.log(__PRETTY_FUNCTION__, MProcess::Section);
 
     const QByteArray &encPass = (gui.radioEntireDisk->isChecked()
         ? gui.textCryptoPass : gui.textCryptoPassCust)->text().toUtf8();
@@ -1060,84 +1101,72 @@ bool PartMan::formatPartitions()
         const QString &useFor = twit->realUseFor();
         if (twit->encrypt) {
             proc.status(tr("Creating encrypted volume: %1").arg(twit->device));
-            if (!luksFormat(twit, encPass)) return false;
+            luksFormat(twit, encPass);
             proc.status();
-            if (!luksOpen(twit, twit->devMapper, encPass)) return false;
+            luksOpen(twit, twit->devMapper, encPass);
         }
         const QString &fmtstatus = tr("Formatting: %1");
         if (useFor == "FORMAT") proc.status(fmtstatus.arg(dev));
         else proc.status(fmtstatus.arg(twit->shownUseFor()));
+        static const char *const msgfail = QT_TR_NOOP("Failed to format partition.");
         if (useFor == "ESP") {
             QStringList cargs({"-F", twit->format.mid(3)});
             if (twit->chkbadblk) cargs.append("-c");
             if (!twit->label.isEmpty()) cargs << "-n" << twit->label.trimmed().left(11);
             cargs.append(dev);
-            if (!proc.exec("mkfs.msdos", cargs)) return false;
+            if (!proc.exec("mkfs.msdos", cargs)) throw msgfail;
             // Sets boot flag and ESP flag.
             const QStringList &devsplit = DeviceItem::split(dev);
             if (!proc.exec("parted", {"-s", "/dev/" + devsplit.at(0),
-                "set", devsplit.at(1), "esp", "on"})) return false;
+                "set", devsplit.at(1), "esp", "on"})) throw msgfail;
         } else if (useFor == "BIOS-GRUB") {
             proc.exec("dd", {"bs=64K", "if=/dev/zero", "of=" + dev});
             const QStringList &devsplit = DeviceItem::split(dev);
             if (!proc.exec("parted", {"-s", "/dev/" + devsplit.at(0),
-                "set", devsplit.at(1), "bios_grub", "on"})) return false;
+                "set", devsplit.at(1), "bios_grub", "on"})) throw msgfail;
         } else if (useFor == "SWAP") {
             QStringList cargs(dev);
             if (!twit->label.isEmpty()) cargs << "-L" << twit->label;
-            if (!proc.exec("/sbin/mkswap", cargs)) return false;
+            if (!proc.exec("/sbin/mkswap", cargs)) throw msgfail;
         } else {
-            if (!formatLinuxPartition(dev, twit->format, twit->chkbadblk, twit->label)) return false;
+            // Transplanted from minstall.cpp and modified to suit.
+            const QString &format = twit->format;
+            QStringList cargs;
+            if (format == "btrfs") {
+                cargs.append("-f");
+                proc.exec("/bin/cp", {"-fp", "/bin/true", "/sbin/fsck.auto"});
+                if (twit->size < 6000000000) {
+                    cargs << "-M" << "-O" << "skinny-metadata"; // Mixed mode (-M)
+                }
+            } else if (format == "xfs" || format == "f2fs") {
+                cargs.append("-f");
+            } else { // jfs, ext2, ext3, ext4
+                if (format == "jfs") cargs.append("-q");
+                else cargs.append("-F");
+                if (twit->chkbadblk) cargs.append("-c");
+            }
+            cargs.append(dev);
+            if (!twit->label.isEmpty()) {
+                if (format == "f2fs") cargs.append("-l");
+                else cargs.append("-L");
+                cargs.append(twit->label);
+            }
+            if (!proc.exec("mkfs." + format, cargs)) throw msgfail;
+            if (format.startsWith("ext")) {
+                proc.exec("/sbin/tune2fs", {"-c0", "-C0", "-i1m", dev}); // ext4 tuning
+            }
         }
     }
     // Prepare subvolumes on all that (are to) contain them.
     for (DeviceItemIterator it(*this); *it; it.next()) {
         if ((*it)->type != DeviceItem::Partition) continue;
-        else if (!prepareSubvolumes(*it)) return false;
+        else prepareSubvolumes(*it);
     }
-
-    return true;
 }
 
-// Transplanted straight from minstall.cpp
-bool PartMan::formatLinuxPartition(const QString &devpath, const QString &format, bool chkBadBlocks, const QString &label)
-{
-    QStringList cargs;
-    if (format == "btrfs") {
-        cargs.append("-f");
-        // btrfs and set up fsck
-        proc.exec("/bin/cp", {"-fp", "/bin/true", "/sbin/fsck.auto"});
-        // set creation options for small drives using btrfs
-        proc.exec("/sbin/sfdisk", {"-s", devpath}, nullptr, true);
-        long long size = proc.readOut().toLongLong();
-        size = size / 1024; // in MiB
-        // if drive is smaller than 6GB, create in mixed mode
-        if (size < 6000) cargs << "-M" << "-O" << "skinny-metadata";
-    } else if (format == "xfs" || format == "f2fs") {
-        cargs.append("-f");
-    } else { // jfs, ext2, ext3, ext4
-        if (format == "jfs") cargs.append("-q");
-        else cargs.append("-F");
-        if (chkBadBlocks) cargs.append("-c");
-    }
-
-    cargs.append(devpath);
-    if (!label.isEmpty()) {
-        if (format == "f2fs") cargs.append("-l");
-        else cargs.append("-L");
-        cargs.append(label);
-    }
-    if (!proc.exec("mkfs." + format, cargs)) return false;
-
-    if (format.startsWith("ext")) {
-        // ext4 tuning
-        proc.exec("/sbin/tune2fs", {"-c0", "-C0", "-i1m", devpath});
-    }
-    return true;
-}
-
-bool PartMan::prepareSubvolumes(DeviceItem *partit)
+void PartMan::prepareSubvolumes(DeviceItem *partit)
 {
+    static const char *const msgfail = QT_TR_NOOP("Failed to prepare subvolumes.");
     const int svcount = partit->childCount();
     QStringList svlist;
     for (int ixi = 0; ixi < svcount; ++ixi) {
@@ -1145,22 +1174,22 @@ bool PartMan::prepareSubvolumes(DeviceItem *partit)
         if (svit->willFormat()) svlist << svit->label;
     }
 
-    if (svlist.isEmpty()) return true;
+    if (svlist.isEmpty()) return;
     proc.status(tr("Preparing subvolumes"));
     svlist.sort(); // This ensures nested subvolumes are created in the right order.
-    bool ok = true;
     mkdir("/mnt/btrfs-scratch", 0755);
-    if (!proc.exec("mount", {"-o", "noatime", partit->mappedDevice(),
-        "/mnt/btrfs-scratch"})) return false;
-    for (const QString &subvol : svlist) {
+    if (!proc.exec("mount", {"-o", "noatime", partit->mappedDevice(), "/mnt/btrfs-scratch"})) {
+        throw msgfail;
+    }
+    bool ok = true;
+    for (const QString &subvol : qAsConst(svlist)) {
         proc.exec("btrfs", {"subvolume", "delete", "/mnt/btrfs-scratch/" + subvol});
-        if (!proc.exec("btrfs", {"subvolume", "create",
-            "/mnt/btrfs-scratch/" + subvol})) ok = false;
+        ok = proc.exec("btrfs", {"subvolume", "create", "/mnt/btrfs-scratch/" + subvol});
         if (!ok) break;
         proc.status();
     }
-    if (!proc.exec("umount", {"/mnt/btrfs-scratch"})) return false;
-    return ok;
+    if (!proc.exec("umount", {"/mnt/btrfs-scratch"})) throw msgfail;
+    if (!ok) throw msgfail;
 }
 
 void PartMan::loadKeyMaterial(const QString &keyfile)
@@ -1169,8 +1198,7 @@ void PartMan::loadKeyMaterial(const QString &keyfile)
     key.load(keyfile.toUtf8().constData(), -1);
     const int keylen = key.length();
     if (keylen > 0) {
-        proc.log(QStringLiteral("Loaded %1-byte key material: ").arg(keylen)
-            + keyfile, MProcess::Standard);
+        proc.log(QStringLiteral("Loaded %1-byte key material: ").arg(keylen) + keyfile);
     }
 }
 
@@ -1235,8 +1263,7 @@ bool PartMan::fixCryptoSetup()
     QTextStream out(&file);
     // Add devices to crypttab.
     for (auto &it : cryptAdd.toStdMap()) {
-        proc.exec("blkid", {"-s", "UUID", "-o", "value", "/dev/" + it.first}, nullptr, true);
-        out << it.second->devMapper << " /dev/disk/by-uuid/" << proc.readOut();
+        out << it.second->devMapper << " /dev/disk/by-uuid/" << it.second->uuid;
         if (noKey || it.first == keyDev) out << " none";
         else {
             if (isNewKey) {
@@ -1262,7 +1289,7 @@ bool PartMan::fixCryptoSetup()
 }
 
 // create /etc/fstab file
-bool PartMan::makeFstab(bool populateMediaMounts)
+bool PartMan::makeFstab()
 {
     QFile file("/mnt/antiX/etc/fstab");
     if (!file.open(QIODevice::WriteOnly)) return false;
@@ -1275,10 +1302,7 @@ bool PartMan::makeFstab(bool populateMediaMounts)
         qDebug() << "Creating fstab entry for:" << it.first << dev;
         // Device ID or UUID
         if (twit->willMap()) out << dev;
-        else {
-            proc.exec("blkid", {"-o", "value", "UUID", "-s", "UUID", dev}, nullptr, true);
-            out << "UUID=" + proc.readOut();
-        }
+        else out << "UUID=" << twit->uuid;
         // Mount point, file system
         proc.exec("blkid", {dev, "-o", "value", "-s", "TYPE"}, nullptr, true);
         const QString &fsfmt = proc.readOut();
@@ -1303,30 +1327,27 @@ bool PartMan::makeFstab(bool populateMediaMounts)
     }
     // EFI System Partition
     if (gui.radioBootESP->isChecked()) {
-        const QString &dev = "/dev/" + gui.comboBoot->currentData().toString();
-        proc.exec("blkid", {"-o", "value", "UUID", "-s", "UUID", dev}, nullptr, true);
-        out << "UUID=" << proc.readOut() << " /boot/efi vfat noatime,dmask=0002,fmask=0113 0 0\n";
+        DeviceItem *devit = findByPath("/dev/" + gui.comboBoot->currentData().toString());
+        if (!devit) return false;
+        out << "UUID=" << devit->uuid << " /boot/efi vfat noatime,dmask=0002,fmask=0113 0 0\n";
     }
     file.close();
-    if (populateMediaMounts) {
-        if (!proc.shell("/sbin/make-fstab -O --install=/mnt/antiX"
-            " --mntpnt=/media")) return false;
-    }
     return true;
 }
 
-bool PartMan::mountPartitions()
+void PartMan::mountPartitions()
 {
-    proc.log(__PRETTY_FUNCTION__);
+    proc.log(__PRETTY_FUNCTION__, MProcess::Section);
     for (auto &it : mounts.toStdMap()) {
         if (it.first.at(0) != '/') continue;
+        static const char *const msgfail = QT_TR_NOOP("Failed to mount partition.");
         const QString point("/mnt/antiX" + it.first);
         const QString &dev = it.second->mappedDevice();
         proc.status(tr("Mounting: %1").arg(dev));
-        if (!proc.mkpath(point)) return false;
+        if (!proc.mkpath(point)) throw msgfail;
         if (it.first == "/boot") {
              // needed to run fsck because sfdisk --part-type can mess up the partition
-            if (!proc.exec("fsck.ext4", {"-y", dev})) return false;
+            if (!proc.exec("fsck.ext4", {"-y", dev})) throw msgfail;
         }
         QStringList opts;
         if (it.second->type == DeviceItem::Subvolume) {
@@ -1339,9 +1360,8 @@ bool PartMan::mountPartitions()
         opts.removeAll("atime");
         opts.removeAll("relatime");
         if (!opts.contains("noatime")) opts << "noatime";
-        if (!proc.exec("/bin/mount", {dev, point, "-o", opts.join(',')})) return false;
+        if (!proc.exec("/bin/mount", {dev, point, "-o", opts.join(',')})) throw msgfail;
     }
-    return true;
 }
 
 void PartMan::unmount()
@@ -1353,7 +1373,7 @@ void PartMan::unmount()
         DeviceItem *twit = mit.value();
         if (!twit) continue;
         if (mit.key().startsWith("SWAP")) {
-            proc.exec("swapoff", {twit->mappedDevice()});
+            proc.shell("grep -q " + twit->mappedDevice() + " /proc/swaps &&  swapoff " + twit->mappedDevice());
         } else if (mit.key().at(0) == '/') {
             proc.exec("/bin/umount", {"-l", "/mnt/antiX" + mit.key()});
         }
@@ -1394,7 +1414,7 @@ int PartMan::isEncrypt(const QString &point)
 {
     int count = 0;
     if (point.isEmpty()) {
-        for (DeviceItem *twit : mounts) {
+        for (DeviceItem *twit : qAsConst(mounts)) {
             if (twit->willEncrypt()) ++count;
         }
     } else if (point == "SWAP") {
@@ -1415,6 +1435,16 @@ DeviceItem *PartMan::findByPath(const QString &devpath) const
     }
     return nullptr;
 }
+DeviceItem *PartMan::findHostDev(const QString &path) const
+{
+    QString spath = path;
+    DeviceItem *devit = nullptr;
+    do {
+        spath = QDir(QFileInfo(spath).absolutePath()).path();
+        devit = mounts.value(spath);
+    } while(!devit && !spath.isEmpty() && spath != '/');
+    return devit;
+}
 
 /*************************\
  * Model View Controller *
@@ -1717,7 +1747,7 @@ DeviceItem::~DeviceItem()
         const int r = parentItem->indexOfChild(this);
         partman->beginRemoveRows(partman->index(parentItem), r, r);
     }
-    for (DeviceItem *cit : children) {
+    for (DeviceItem *cit : qAsConst(children)) {
         cit->partman = nullptr; // Stop unnecessary signals.
         cit->parentItem = nullptr; // Stop double deletes.
         delete cit;
@@ -1736,7 +1766,7 @@ void DeviceItem::clear()
 {
     const int chcount = children.count();
     if (partman && chcount > 0) partman->beginRemoveRows(partman->index(this), 0, chcount - 1);
-    for (DeviceItem *cit : children) {
+    for (DeviceItem *cit : qAsConst(children)) {
         cit->partman = nullptr; // Stop unnecessary signals.
         cit->parentItem = nullptr; // Stop double deletes.
         delete cit;
@@ -1776,7 +1806,7 @@ void DeviceItem::sortChildren()
     };
     std::sort(children.begin(), children.end(), cmp);
     if (partman) {
-        for (DeviceItem *c : children) partman->notifyChange(c);
+        for (DeviceItem *c : qAsConst(children)) partman->notifyChange(c);
     }
 }
 /* Helpers */
@@ -1792,9 +1822,9 @@ QString DeviceItem::shownUseFor() const
 {
     const QString &use = realUseFor();
     if (use == "/") return "/ (root)";
-    else if (use == "ESP") return qApp->tr("EFI System Partition");
-    else if (use == "SWAP") return qApp->tr("swap space");
-    else if (use == "FORMAT") return qApp->tr("format only");
+    else if (use == "ESP") return tr("EFI System Partition");
+    else if (use == "SWAP") return tr("swap space");
+    else if (use == "FORMAT") return tr("format only");
     return use;
 }
 void DeviceItem::setActive(bool boot)
@@ -1885,10 +1915,10 @@ QStringList DeviceItem::allowedUsesFor(bool real) const
 QStringList DeviceItem::allowedFormats() const
 {
     QStringList list;
-    bool allowPreserve = false, selPreserve = false;
+    const QString &use = realUseFor();
+    bool allowPreserve = false;
     if (type == Subvolume) list.append("CREATE");
     else if (isVolume()) {
-        const QString &use = realUseFor();
         if (use.isEmpty()) return QStringList();
         else if (use == "/boot") list.append("ext4");
         else if (use == "BIOS-GRUB") list.append("GRUB");
@@ -1896,33 +1926,32 @@ QStringList DeviceItem::allowedFormats() const
             list.append("FAT32");
             if (size <= 4294901760) list.append("FAT16");
             if (size <= 33553920) list.append("FAT12");
-            selPreserve = allowPreserve = (list.contains(curFormat, Qt::CaseInsensitive)
+            allowPreserve = (list.contains(curFormat, Qt::CaseInsensitive)
                 || !curFormat.compare("VFAT", Qt::CaseInsensitive));
         } else if (use == "SWAP") {
             list.append("SWAP");
-            selPreserve = allowPreserve = list.contains(curFormat, Qt::CaseInsensitive);
+            allowPreserve = list.contains(curFormat, Qt::CaseInsensitive);
         } else {
             list << "ext4" << "ext3" << "ext2";
             list << "f2fs" << "jfs" << "xfs" << "btrfs";
             if (use != "FORMAT") allowPreserve = list.contains(curFormat, Qt::CaseInsensitive);
-            if (use == "/home") selPreserve = allowPreserve;
         }
     }
     if (encrypt) allowPreserve = false;
     if (allowPreserve) {
-        if (selPreserve) list.prepend("PRESERVE");
+        if (use != "/") list.prepend("PRESERVE");
         else list.append("PRESERVE");
     }
     return list;
 }
 QString DeviceItem::shownFormat(const QString &fmt) const
 {
-    if (fmt == "CREATE") return qApp->tr("Create");
+    if (fmt == "CREATE") return tr("Create");
     else if (fmt != "PRESERVE") return fmt;
     else {
-        if (type == Subvolume) return qApp->tr("Preserve");
-        else if (realUseFor() != "/") return qApp->tr("Preserve (%1)").arg(curFormat);
-        else return qApp->tr("Preserve /home (%1)").arg(curFormat);
+        if (type == Subvolume) return tr("Preserve");
+        else if (realUseFor() != "/") return tr("Preserve (%1)").arg(curFormat);
+        else return tr("Preserve /home (%1)").arg(curFormat);
     }
 }
 bool DeviceItem::canMount() const
@@ -2022,8 +2051,8 @@ void DeviceItem::autoFill(unsigned int changed)
                 }
                 options.clear();
                 if (!flags.rotational) {
-                    const bool btrfs = (format=="btrfs" || type==Subvolume);
-                    if (discgran && (format=="ext4" || format=="xfs")) options = "discard,";
+                    const bool btrfs = (format == "btrfs" || type == Subvolume);
+                    if (discgran && (format == "ext4" || format == "xfs")) options = "discard,";
                     else if (discgran && btrfs) options = "ssd,discard=async,";
                     else if (btrfs) options = "ssd,";
                 }
@@ -2064,22 +2093,11 @@ int DeviceItem::layoutDefault(int rootPercent, bool crypto, bool updateTree)
     const int bootMinMB = static_cast<int>(partman->bootSpaceNeeded / 1048576);
     if (!crypto) rootMinMB += bootMinMB;
     else {
-        int bootFormatSize = 512;
+        int bootFormatSize = 1024;
         if (bootFormatSize < bootMinMB) bootFormatSize = static_cast<int>(partman->bootSpaceNeeded);
         if (updateTree) addPart(bootFormatSize, "boot", crypto);
         rootFormatSize -= bootFormatSize;
     }
-    // Swap space.
-    int swapFormatSize = rootFormatSize-rootMinMB;
-    struct sysinfo sinfo;
-    if (sysinfo(&sinfo) != 0) sinfo.totalram = 2048;
-    else sinfo.totalram = (sinfo.totalram / (1048576 * 2)) * 3; // 1.5xRAM
-    sinfo.totalram /= 128; ++sinfo.totalram; sinfo.totalram *= 128; // Multiple of 128MB
-    if (swapFormatSize > static_cast<int>(sinfo.totalram)) swapFormatSize = static_cast<int>(sinfo.totalram);
-    int swapMaxMB = rootFormatSize / (20 * 128); ++swapMaxMB; swapMaxMB *= 128; // 5% root
-    if (swapMaxMB > 8192) swapMaxMB = 8192; // 8GB cap for the whole calculation.
-    if (swapFormatSize > swapMaxMB) swapFormatSize = swapMaxMB;
-    rootFormatSize -= swapFormatSize;
     // Home
     int homeFormatSize = rootFormatSize;
     rootFormatSize = (rootFormatSize * rootPercent) / 100;
@@ -2088,7 +2106,6 @@ int DeviceItem::layoutDefault(int rootPercent, bool crypto, bool updateTree)
 
     if (updateTree) {
         addPart(rootFormatSize, "root", crypto);
-        if (swapFormatSize>0) addPart(swapFormatSize, "swap", crypto);
         if (homeFormatSize>0) addPart(homeFormatSize, "home", crypto);
         labelParts();
         driveAutoSetActive();
